import 'package:flutter/material.dart';
import 'package:flutter_gadgets/flutter_gadgets.dart';

/// A void function used to initialize the internal state of a given view model.
typedef void ViewModelInitState<T>(T viewModel);

/// A void function used to dispose the internal state of a given view model.
typedef void ViewModelDispose<T>(T viewModel);

/// A [SatefulWidget] that encapsulate the state of a view model.
/// During initialization, the view model is automatically registered
/// into the [ObservableModel] and the optional initStateCallback is executed.
/// When the widget is disposed, the view model is automatically removed
/// from the [ObservableModel] and the optional disposeCallback is executed.
///
/// The disposeCallback can be used to manually dispose stateful components of
/// the given view model (like a [TextEditingController]).
class ViewModelGadget<T> extends StatefulWidget {
  ViewModelGadget(
    this.viewModel, {
    this.initStateCallback,
    this.disposeCallback,
    @required this.child,
  });

  /// An object representing the internal state of a view.
  final T viewModel;

  /// An optional [ViewModelInitState] callback executed during the [ViewModelGadget] initState.
  final ViewModelInitState<T> initStateCallback;

  /// An optional [ViewModelDispose] callback executed during the [ViewModelGadget] dispose.
  final ViewModelDispose<T> disposeCallback;

  /// The view encapsulated by the [ViewModelGadget].
  final Widget child;

  @override
  _ViewModelGadgetState<T> createState() => _ViewModelGadgetState<T>();
}

class _ViewModelGadgetState<T> extends State<ViewModelGadget<T>> {
  @override
  void initState() {
    super.initState();
    if (widget.initStateCallback != null)
      widget.initStateCallback(widget.viewModel);
    ServiceContainer.instance.observableModel.putBean<T>(
      widget.viewModel,
      notify: false,
    );
  }

  @override
  void dispose() {
    if (widget.disposeCallback != null)
      widget.disposeCallback(widget.viewModel);
    ServiceContainer.instance.observableModel.removeBean<T>();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return widget.child;
  }
}
